#include "road_model.h"

#include <algorithm>
#include <iostream>
#include <limits>
#include <list>
#include <unordered_map>
#include <utility>

#include "adapter_output.h"
#include "boundary.h"
#include "boundary_builder_observer.h"
#include "corridor_builder_observer.h"
#include "file.h"
#include "glog/logging.h"
#include "ilc_corridor_builder_observer.h"
#include "lane.h"
#include "macros.h"
#include "object_assignment_builder_observer.h"
#include "road_info_adapter.h"
#include "road_info_provider_observer.h"
#include "task_factory.h"
using namespace std;

namespace rm {

using rm::common::ErrorCode;
using rm::common::Status;

RoadModel::~RoadModel() {}

std::string RoadModel::Name() const { return "road_model"; }

common::Status RoadModel::ObserverInit() {
  config_ =
      shared_ptr<RoadModelSubject>(new RoadModelSubject("RoadModelSubject"));
  std::shared_ptr<Observer> observerA = shared_ptr<Observer>(
      new BoundaryBuilderObserver(config_)); //设置观察者A 观察具体目标A
  std::shared_ptr<Observer> observerB = shared_ptr<Observer>(
      new CorridorBuilderObserver(config_)); //设置观察者B 观察具体目标A
  std::shared_ptr<Observer> observerC = shared_ptr<Observer>(
      new ObjectAssignmentBuilderObserver(config_)); //设置观察者B 观察具体目标A
  std::shared_ptr<Observer> observerD = shared_ptr<Observer>(
      new RoadInfoProviderObserver(config_)); // road info provider
  std::shared_ptr<Observer> observerE = shared_ptr<Observer>(
      new ILCCorridorBuilderObserver(config_)); // ILC corridor
  observer_[observerA->getDynamicConfig()->config_type()] = observerA;
  observer_[observerB->getDynamicConfig()->config_type()] = observerB;
  observer_[observerC->getDynamicConfig()->config_type()] = observerC;
  observer_[observerD->getDynamicConfig()->config_type()] = observerD;
  // FIXME ilc corridor builder can not be constructed
  // observer_[observerE->getDynamicConfig()->config_type()] = observerE;
  // observer_.push_back(observerA);
  // observer_.push_back(observerB);
  // observer_.push_back(observerC);
  // observer_.push_back(observerD);

  for (auto &observer : observer_) {
    config_->addObserver(observer.second.get()); //具体目标A添加观察者A
  }
  config_->Start();
  return Status::OK();
}

Status RoadModel::Init() {
  // 1.config observer 初始化
  ObserverInit();

  // 2.注册task
  TaskFactory::Init();

  // 3.根据observer创建task
  
  for (auto &observer : observer_) {
    std::unique_ptr<Task> ptr =
        TaskFactory::CreateTask(observer.second, injector_);
    if (ptr == nullptr)
      continue;
    task_list_.push_back(ptr.get());
    tasks_[observer.first] = std::move(ptr);
  }
  // 4.车道提供器初始化
  road_info_provider_ = std::unique_ptr<RoadInfoAdapter>(new RoadInfoAdapter(
      observer_[DynamicConfig::ConfigType::ROAD_INFO_PROVIDER]));

  // counters_ = static_cast<uint16_t>(0);
  return Status::OK();
}

Status RoadModel::InitFrame(const uint32_t sequence_num) {
  /**********frame重置**********/
  
  frame_.reset(new Frame(
      sequence_num, local_view_,
      road_info_provider_.get())); // reset：释放原来指向的空间 重新指向新的空间
    if (frame_ == nullptr) {
    return Status(ErrorCode::RM_ERROR, "Fail to init frame: nullptr.");
  }
  

  std::list<Lane> lanes; //存储输入信息预处理后的corridor 不一定是corridor
                         //可能是laneinfo  按后frame初始化时创建corridorinfo
  if (!road_info_provider_->GetLanes(&lanes)) { //获取Lanes用于创建laneInfo
    const std::string msg = "Failed to create reference line";
    // DLOG(ERROR) << msg;
    return Status(ErrorCode::RM_ERROR, msg);
  }
  auto status = frame_->Init(lanes);
  
  if (!frame_->traffic_road_input_status()) {
    // traffic road input is invalid
    ResetInjector();
    BoundaryBuilder::ResetBoundaryBuilder();
    if (!frame_->objects_input_status()) {
      // objects input is invalid
    //   DLOG(WARNING) << "failed to init frame: traffic road and objects input "
    //                    "both invalid";
      return Status(ErrorCode::RM_INIT_ERROR, "All input data not valid");
    }
  }
  if (!status.ok()) {
    // DLOG(WARNING) << "failed to init frame:" << status.ToString();
    return status;
  }
    return Status::OK();
}

void RoadModel::RunOnce(const LocalView &local_view,
                        dpc::RoadInfo *const ptr_roadinfo,
                        dpc::RMSMMessage *const ptr_rm_sm_msg,
                        dpc::TargetsInfo *const ptr_targetsinfo) {
  
  local_view_ = local_view;
  //更新下输入信息到road_info_provider
  VehicleStateProvider::Instance()->Update(
      *local_view_.vehicle_state/*, *local_view_.sas_10ms_pdu00,
      *local_view_.eps_10ms_pdu00, *local_view_.sm_info*/);
  road_info_provider_->Update(local_view);
  
  RelayCemInputStatus(ptr_rm_sm_msg);

  Status status = InitFrame(static_cast<uint32_t>(seq_num_++));
  if (!status.ok()) { //初始化frame
                      // traffic road and objects both invalid
    // TODO(ningmeng):输出并向下游发布无效的数据
    return;
  }
  status = PreProcess(); // *
  if (!status.ok()) {
    return;
  }
  

  status = Process(); // *
  if (!status.ok()) {
    // TODO(ningmeng):输出并向下游发布无效的数据
    return;
  }
  status = PostProcess(); // *
  
  status = UpdateOutputInfo(ptr_roadinfo, ptr_rm_sm_msg, ptr_targetsinfo);
}
Status RoadModel::Process() {
  // TODO(ningmeng): 对roadinfo的每条lane进行遍历进行process
  if (frame_->traffic_road_input_status()) {
    for (auto &lane_info : *frame_->mutable_lane_info()) {
      // DLOG(INFO) << PRINT(+lane_info.lane().lane_id());
      Status status = ProcessOnLane(&lane_info);
    }
  }
  return Status::OK();
}
Status RoadModel::ProcessOnLane(LaneInfo *lane_info) {
  auto ret = Status::OK();
  for (auto *task : task_list_) {
    // DLOG(INFO) << "task is : " << task->Name();

    ret = task->Execute(frame_.get(), lane_info);

    if (!ret.ok()) {
    //   DLOG(ERROR) << "Failed to run tasks[" << task->Name()
    //               << "], Error message: " << ret.error_message();
      break;
    }
  }
  return Status::OK();
}
Status
RoadModel::UpdateOutputInfo(dpc::RoadInfo *ptr_roadinfo,
                            dpc::RMSMMessage *ptr_rm_sm_msg,
                            dpc::TargetsInfo *ptr_targetsinfo) {
  // Header output
  ptr_roadinfo->fusion_timestamp =
      road_info_provider_->mutable_traffic_road_cem_header()->timestamp();
  // uint16_t rm_cyclecounter = RunCounter();
  // ptr_roadinfo->rm_header.cyclecounter = rm_cyclecounter;
  // ptr_rm_sm_msg->rm_header.cyclecounter = rm_cyclecounter;
  // ptr_targetsinfo->cyclecounter = rm_cyclecounter;
  // ptr_roadinfo->rm_header.software_version = rm::SOFTWAREVERSION;
  //首阶段只对Ego_Lane作处理,将frame的信息赋值到输出的message结构体中
  // vector<LanePosition> position_data_ = {
  //     LanePosition::LEFT6,    LanePosition::LEFT5,  LanePosition::LEFT4,
  //     LanePosition::LEFT3,    LanePosition::LEFT2,  LanePosition::LEFT1,
  //     LanePosition::EGO_LANE, LanePosition::RIGHT1, LanePosition::RIGHT2,
  //     LanePosition::RIGHT3,   LanePosition::RIGHT4, LanePosition::RIGHT5,
  //     LanePosition::RIGHT6};
  // 1.存储lane list信息
  uint32_t left_line_id, right_line_id, ll_line_id, rr_line_id;
  uint8_t *ego_lane_id_prev_ptr = injector_->mutable_ego_lane_id_last_cycle();
  uint8_t lane_line_crossed_temp = static_cast<uint8_t>(0);
  bool flag = true;
  bool hdmap_flag =
      (frame_->road_info_provider()->traffic_road_sensor_status() ==
       rm::SensorStatus::FVCM_AND_HDM_VALID);
  //   road_info_provider_->mutable_lanes()->back().
  for (auto &lane_info : *frame_->mutable_lane_info()) {
    // DLOG(INFO) << PRINT(+lane_info.mutable_lane()->lane_id());
    // common information no matter boundary_pair_type of lane is valid or not
    dpc::Lane lane;
    lane.start_id = lane_info.mutable_lane()->lane_id();
    lane.end_id = lane_info.mutable_lane()->lane_id();
    lane.lane_type =
        static_cast<uint8_t>(lane_info.mutable_lane()->lane_type());

    lane.start_range = lane_info.mutable_lane()->start_offset();
    lane.end_range = lane_info.mutable_lane()->end_offset();
    lane.is_start_out_of_range =
        lane_info.mutable_lane()->is_start_out_of_range();
    lane.is_end_out_of_range = lane_info.mutable_lane()->is_end_out_of_range();
    output::MaxSpeedLimitAdapter(lane, lane_info.lane());

    if (hdmap_flag == false) {
      for (int i = 0; i < position_data_.size(); i++) {
        if (lane_info.mutable_lane()->lane_position() != position_data_[i])
          continue;
        if (i - 1 >= 0) {
          Lane *left_lane_ptr =
              frame_->FindLanePtrFromLanePosition(position_data_[i - 1]);
          if (left_lane_ptr != nullptr) {
            dpc::NeighborLane temp;
            temp.start_range = left_lane_ptr->start_offset();
            temp.end_range = left_lane_ptr->end_offset();
            temp.lane_id = left_lane_ptr->lane_id();
            lane.left_neighbor_lane.emplace_back(temp);
          }
        }
        if (i + 1 < position_data_.size()) {
          Lane *right_lane_ptr =
              frame_->FindLanePtrFromLanePosition(position_data_[i + 1]);
          if (right_lane_ptr != nullptr) {
            dpc::NeighborLane temp;
            temp.start_range = right_lane_ptr->start_offset();
            temp.end_range = right_lane_ptr->end_offset();
            temp.lane_id = right_lane_ptr->lane_id();
            lane.right_neighbor_lane.emplace_back(temp);
          }
        }
      }
    } else {
      WPExtendLane wp_extend_lane =
          road_info_provider_->GetExtendLaneFromId(lane.start_id);
      if (wp_extend_lane.lane_id() != 0) {
        for (const auto &left_lane : wp_extend_lane.left_extend_lane()) {
          if (left_lane.lane_end_offset <= static_cast<float>(0.0))
            continue;
          dpc::NeighborLane left_adj_lane;
          left_adj_lane.lane_id = left_lane.lane_id;
          left_adj_lane.start_range = left_lane.lane_offset;
          left_adj_lane.end_range = left_lane.lane_end_offset;
          lane.left_neighbor_lane.emplace_back(left_adj_lane);
        }
        for (const auto &right_lane : wp_extend_lane.right_extend_lane()) {
          if (right_lane.lane_end_offset <= static_cast<float>(0.0))
            continue;
          dpc::NeighborLane right_adj_lane;
          right_adj_lane.lane_id = right_lane.lane_id;
          right_adj_lane.start_range = right_lane.lane_offset;
          right_adj_lane.end_range = right_lane.lane_end_offset;
          lane.right_neighbor_lane.emplace_back(right_adj_lane);
        }
      }
    }

    std::list<uint32_t> left_lane_marker_id =
        lane_info.mutable_lane()->left_lane_marker_id();
    std::list<uint32_t> right_lane_marker_id =
        lane_info.mutable_lane()->right_lane_marker_id();
    // TODO(ningmeng）： 暂时把lanemarker的id视为boundary的id
    // 因为cem的lanemarkersegmentid是不唯一的
    lane.left_lane_marker_id_list.assign(left_lane_marker_id.begin(),
                                         left_lane_marker_id.end());
    lane.right_lane_marker_id_list.assign(right_lane_marker_id.begin(),
                                          right_lane_marker_id.end());

    // if lane marker segement 0-50 invalid on both side, but exist in cem lane
    // list, output default corridor
    if (lane_info.boundary_pair_type() == BoundaryPairType::BOTH_INVALID) {
      dpc::Corridor cor;
      output::DefaultCorridor2Corridor(cor);
      lane.corridor_list.emplace_back(std::move(cor));
      ptr_roadinfo->lane_list.push_back(lane);
      continue;
    }
    //仅输出 egolane
    if (lane_info.mutable_lane()->lane_position() == LanePosition::EGO_LANE) {
      ptr_roadinfo->car_position.ego_lane_id =
          lane_info.mutable_lane()->lane_id();
      if (*ego_lane_id_prev_ptr == static_cast<uint8_t>(0)) {
        // initial case where no lane is recevied in last cycle
        lane_line_crossed_temp =
            static_cast<uint8_t>(LaneLineCrossedType::NONE);
      } else if (*ego_lane_id_prev_ptr !=
                 ptr_roadinfo->car_position.ego_lane_id) {
        // case where the ego lane id changes
        if (*ego_lane_id_prev_ptr ==
            frame_->FindLaneIdFromLanePosition(LanePosition::LEFT1)) {
          lane_line_crossed_temp =
              static_cast<uint8_t>(LaneLineCrossedType::CROSSED_RIGHT_LINE);
        } else if (*ego_lane_id_prev_ptr ==
                   frame_->FindLaneIdFromLanePosition(LanePosition::RIGHT1)) {
          lane_line_crossed_temp =
              static_cast<uint8_t>(LaneLineCrossedType::CROSSED_LEFT_LINE);
        } else {
          lane_line_crossed_temp =
              static_cast<uint8_t>(LaneLineCrossedType::NONE);
        }
      } else {
        // case where ego lane id remains unchanged
        lane_line_crossed_temp =
            static_cast<uint8_t>(LaneLineCrossedType::NONE);
      }
      *ego_lane_id_prev_ptr = ptr_roadinfo->car_position.ego_lane_id;
      ptr_roadinfo->car_position.lane_line_crossed =
          line_crossed_delay_.RunDelay(lane_line_crossed_temp);
      ptr_rm_sm_msg->rm_lne_crss = ptr_roadinfo->car_position.lane_line_crossed;
      frame_->set_lane_line_crossed(lane_line_crossed_temp);
      // bool is_left_marker_from_hdmap =
      //     (lane_info.boundary().first.sensor_source() &
      //      static_cast<uint16_t>(SensorSource::HDMAP));
      // bool is_right_marker_from_hdmap =
      //     (lane_info.boundary().second.sensor_source() &
      //      static_cast<uint16_t>(SensorSource::HDMAP));
      // hdmap_flag = (is_left_marker_from_hdmap || is_right_marker_from_hdmap)
      //                  ? true
      //                  : false;
      // road_info_provider_->GetLaneMarkerFromId()
      rm::ControlWayPoints *control_waypoints_ptr =
          frame_->road_info_provider()->mutable_way_points();
      dpc::LaneInfo ego_lane_info;
      ego_lane_info.lane_id = lane_info.mutable_lane()->lane_id();
      ego_lane_info.start_offset = lane_info.mutable_lane()->start_offset();
      ego_lane_info.end_offset = lane_info.mutable_lane()->end_offset();
      ego_lane_info.lane_merge_distance =
          lane_info.mutable_lane()->lane_merge_distance();
      output::ReferenceLineGeometeryAdapter(ego_lane_info, &lane_info);
      output::ODDInfoAdapter(ego_lane_info,
                             control_waypoints_ptr->FindWPExtendLaneByLaneId(
                                 ego_lane_info.lane_id));
      ego_lane_info.lane_type =
          static_cast<uint8_t>(lane_info.mutable_lane()->lane_type());
      if (lane_info.mutable_lane()->lane_type() == LaneType::ACCLERATION_LANE ||
          lane_info.mutable_lane()->lane_type() == LaneType::ENTRY) {
        if (lane_info.mutable_lane()->speed_limit().first > 90.0) {
          ego_lane_info.max_speed = 90.0;
          ego_lane_info.min_speed = 90.0;
        } else {
          ego_lane_info.max_speed =
              lane_info.mutable_lane()->speed_limit().first - 10.0;
//          ego_lane_info.min_speed = ego_lane_info.min_speed;
        }

      } else {
        ego_lane_info.max_speed = lane_info.mutable_lane()->speed_limit().first;
        ego_lane_info.min_speed =
            lane_info.mutable_lane()->speed_limit().second;
      }

      ptr_rm_sm_msg->ego_lane.at(0) = ego_lane_info;

      for (int i = 0; i < position_data_.size(); i++) {
        if (i < 6) {
          LaneInfo *left_lane_info_ptr =
              frame_->FindLaneInfoPtrFromLanePosition(position_data_[i]);
          if (left_lane_info_ptr != nullptr) {
            output::Lane2LaneInfo(
                left_lane_info_ptr,
                ptr_rm_sm_msg->left_lane.at(
                    static_cast<uint8_t>(position_data_[i]) - 2U));
            output::ODDInfoAdapter(
                ptr_rm_sm_msg->left_lane.at(
                    static_cast<uint8_t>(position_data_[i]) - 2U),
                control_waypoints_ptr->FindWPExtendLaneByLaneId(
                    left_lane_info_ptr->mutable_lane()->lane_id()));
          }
        }
        if (i > 6) {
          LaneInfo *right_lane_info_ptr =
              frame_->FindLaneInfoPtrFromLanePosition(position_data_[i]);
          if (right_lane_info_ptr != nullptr) {
            output::Lane2LaneInfo(
                right_lane_info_ptr,
                ptr_rm_sm_msg->right_lane.at(
                    static_cast<uint8_t>(position_data_[i]) - 8U));
            output::ODDInfoAdapter(
                ptr_rm_sm_msg->right_lane.at(
                    static_cast<uint8_t>(position_data_[i]) - 8U),
                control_waypoints_ptr->FindWPExtendLaneByLaneId(
                    right_lane_info_ptr->mutable_lane()->lane_id()));
          }
        }
      }
    }
    for (auto lane_seg_rev_iter =
             lane_info.mutable_lane()->mutable_back_lane_segment()->rbegin();
         lane_seg_rev_iter !=
         lane_info.mutable_lane()->mutable_back_lane_segment()->rend();
         lane_seg_rev_iter++) {
      for (const auto &sample_point :
           lane_seg_rev_iter->sampling_point_list()) {
        dpc::SamplingPoint pubs_sampling_point;
        output::SamplingPoint2SamplingPoint(sample_point, pubs_sampling_point);
        lane.sampling_point_list.emplace_back(std::move(pubs_sampling_point));
      }
    }
    for (const auto &lane_segment :
         lane_info.mutable_lane()->front_lane_segment()) {
      for (const auto &sample_point : lane_segment.sampling_point_list()) {
        dpc::SamplingPoint pubs_sampling_point;
        output::SamplingPoint2SamplingPoint(sample_point, pubs_sampling_point);
        lane.sampling_point_list.emplace_back(std::move(pubs_sampling_point));
      }
    }

    // corridor
    // DLOG(INFO) << PRINT(lane_info.mutable_corridor()->size());
    bool is_ego_lane =
        (lane_info.mutable_lane()->lane_position() == LanePosition::EGO_LANE);
    uint8_t idx_counter = 0U;
    for (auto &corridor : *lane_info.mutable_corridor()) {
      if (corridor.is_first_seg() ||
          corridor.feature_tag() == rm::CorridorFeatureTag::LINEDEVIATE ||
          corridor.feature_tag() == rm::CorridorFeatureTag::CARFOLLOW) {
        // only output the first corridor seg and the ilc corridors
        dpc::Corridor cor;
        output::Corridor2Corridor(corridor, cor);
        lane.corridor_list.emplace_back(std::move(cor));
        if (is_ego_lane) {
          ptr_rm_sm_msg->rm_crrdr_id.at(idx_counter) = corridor.corridor_id();
          ptr_rm_sm_msg->rm_featuretag.at(idx_counter) =
              static_cast<uint8_t>(corridor.feature_tag());
          idx_counter++;
        }
      }
      // break;
    }
    // DLOG(INFO) << PRINT(lane_info.mutable_back_corridor()->size());
    for (auto &corridor : *lane_info.mutable_back_corridor()) {
      if (corridor.is_first_seg()) {
        // only output the first back corridor seg
        dpc::Corridor cor;
        output::Corridor2Corridor(corridor, cor);
        lane.back_corridor_list.emplace_back(std::move(cor));
      }
    }
    ptr_roadinfo->lane_list.push_back(lane);

    //取出sm需要的左右车道线id
    if (flag &&
        lane_info.mutable_lane()->lane_position() == LanePosition::EGO_LANE) {
      left_line_id =
          left_lane_marker_id.empty() ? 0 : left_lane_marker_id.front();
      right_line_id =
          right_lane_marker_id.empty() ? 0 : right_lane_marker_id.front();
      ptr_rm_sm_msg->rm_lanewidth = lane_info.mutable_lane()->lane_width();
      flag = !flag;
    }
    if (lane_info.mutable_lane()->lane_position() == LanePosition::LEFT1) {
      ll_line_id =
          left_lane_marker_id.empty() ? 0 : left_lane_marker_id.front();
    }
    if (lane_info.mutable_lane()->lane_position() == LanePosition::RIGHT1) {
      rr_line_id =
          right_lane_marker_id.empty() ? 0 : right_lane_marker_id.front();
    }
  }

  // 2.存储boundary_list 信息
  std::vector<LaneMarker> *lane_marker_ptr =
      road_info_provider_->mutable_lane_markers();
  for (auto it = lane_marker_ptr->begin(); it != lane_marker_ptr->end(); it++) {
    if (it->mutable_front_lane_marker_segment()->empty())
      continue;
    // TODO(ningmeng)暂时先取front的一段即可
    dpc::LaneMarker lane_marker;
    output::LaneMarker2LaneMarker(*it, lane_marker);
    ptr_roadinfo->lane_marker_list.emplace_back(std::move(lane_marker));

    //将自车道的左右边线赋值
    if (left_line_id == it->lane_marker_id()) {
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_id = left_line_id;
      rm::PolynomLine *l_line_curve_ptr =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->mutable_curve();
      rm::LineGeometry *l_line_geometry_ptr =
          it->mutable_front_lane_marker_segment()->front().mutable_geometry();
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].c0_llne =
          l_line_curve_ptr->Coef(0);
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].c1_llne =
          l_line_curve_ptr->Coef(1);
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].c2_llne =
          l_line_curve_ptr->Coef(2);
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].c3_llne =
          l_line_curve_ptr->Coef(3);
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].start_rnge_llne =
          l_line_curve_ptr->start_range();
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].end_rnge_llne =
          l_line_curve_ptr->end_range();
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].probability_llne =
          static_cast<float>(l_line_geometry_ptr->valid());
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_geometry[0].valid_llne =
          l_line_geometry_ptr->valid();
      ptr_targetsinfo->left_lane_marker.lane_mkr_c0 = l_line_curve_ptr->Coef(0);
      ptr_targetsinfo->left_lane_marker.lane_mkr_c1 = l_line_curve_ptr->Coef(1);
      // ptr_targetsinfo->left_lane_marker.lane_mkr_c2 =
      // l_line_curve_ptr->Coef(2);
      // ptr_targetsinfo->left_lane_marker.lane_mkr_c3 =
      // l_line_curve_ptr->Coef(3);
      ptr_targetsinfo->left_lane_marker.lane_mkr_c2 =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->hdm_c2();
      ptr_targetsinfo->left_lane_marker.lane_mkr_c3 =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->hdm_c3();
      ptr_targetsinfo->left_lane_marker.lane_mrk_viewrange =
          l_line_curve_ptr->end_range();
      ptr_targetsinfo->left_lane_marker.lane_mkr_existnc_problty =
          static_cast<float>(l_line_geometry_ptr->valid());
      ptr_targetsinfo->left_lane_marker.lane_mkr_typ =
          static_cast<uint8_t>(it->lane_marker_type());
      output::SampingPointTrpLongAdapter(
          ptr_targetsinfo->left_lane_marker.sampling_points,
          it->mutable_front_lane_marker_segment(),
          frame_->road_info_provider()->desired_lane_id(),
          frame_->road_info_provider()->special_ego_lane());
      // ptr_rm_sm_msg->left_line[0].rm_egolftlne_color =
      //     static_cast<uint8_t>(it->mutable_front_lane_marker_segment()->front().color());
      // ptr_rm_sm_msg->left_line[0].rm_egolftlne_type =
      //     static_cast<uint8_t>(it->mutable_front_lane_marker_segment()->front().type());
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_color =
          static_cast<uint8_t>(it->lane_marker_color());
      rm::LineType cur_ln_type = it->lane_marker_type();
      ptr_rm_sm_msg->left_line[0].rm_egolftlne_type =
          static_cast<uint8_t>(cur_ln_type);
      // FIXME rm_lftlnesnsrsource shall be uint16
      ptr_rm_sm_msg->left_line[0].rm_lftlnesnsrsource = static_cast<uint8_t>(
          it->mutable_front_lane_marker_segment()->front().sensor_source());
      ptr_rm_sm_msg->rm_dsblrsn[2] = static_cast<uint8_t>(
          output::MeasurementStateRemap(it->mutable_front_lane_marker_segment()
                                            ->front()
                                            .mutable_geometry()
                                            ->measurement_state(),
                                        cur_ln_type));
      // static_cast<uint32_t>(it->mutable_front_lane_marker_segment()
      //                           ->front()
      //                           .mutable_geometry()
      //                           ->measurement_state());
    }
    if (right_line_id == it->lane_marker_id()) {
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_id = right_line_id;
      rm::PolynomLine *r_line_curve_ptr =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->mutable_curve();
      rm::LineGeometry *r_line_geometry_ptr =
          it->mutable_front_lane_marker_segment()->front().mutable_geometry();
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].c0_rlne =
          r_line_curve_ptr->Coef(0);
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].c1_rlne =
          r_line_curve_ptr->Coef(1);
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].c2_rlne =
          r_line_curve_ptr->Coef(2);
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].c3_rlne =
          r_line_curve_ptr->Coef(3);
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].start_rnge_rlne =
          r_line_curve_ptr->start_range();
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].end_rnge_rlne =
          r_line_curve_ptr->end_range();
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].probability_rlne =
          static_cast<float>(r_line_geometry_ptr->valid());
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_geometry[0].valid_rlne =
          r_line_geometry_ptr->valid();
      ptr_targetsinfo->right_lane_marker.lane_mkr_c0 =
          r_line_curve_ptr->Coef(0);
      ptr_targetsinfo->right_lane_marker.lane_mkr_c1 =
          r_line_curve_ptr->Coef(1);
      // ptr_targetsinfo->right_lane_marker.lane_mkr_c2 =
      //     r_line_curve_ptr->Coef(2);
      // ptr_targetsinfo->right_lane_marker.lane_mkr_c3 =
      //     r_line_curve_ptr->Coef(3);
      ptr_targetsinfo->right_lane_marker.lane_mkr_c2 =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->hdm_c2();
      ptr_targetsinfo->right_lane_marker.lane_mkr_c3 =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->hdm_c3();
      ptr_targetsinfo->right_lane_marker.lane_mrk_viewrange =
          r_line_curve_ptr->end_range();
      ptr_targetsinfo->right_lane_marker.lane_mkr_existnc_problty =
          static_cast<float>(r_line_geometry_ptr->valid());
      ptr_targetsinfo->right_lane_marker.lane_mkr_typ =
          static_cast<uint8_t>(it->lane_marker_type());
      output::SampingPointTrpLongAdapter(
          ptr_targetsinfo->right_lane_marker.sampling_points,
          it->mutable_front_lane_marker_segment(),
          frame_->road_info_provider()->desired_lane_id(),
          frame_->road_info_provider()->special_ego_lane());
      // ptr_rm_sm_msg->right_line[0].rm_egorgtlne_color =
      //     static_cast<uint8_t>(it->front_lane_marker_segment().front().color());
      // ptr_rm_sm_msg->right_line[0].rm_egorgtlne_type =
      //     static_cast<uint8_t>(it->front_lane_marker_segment().front().type());
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_color =
          static_cast<uint8_t>(it->lane_marker_color());
      rm::LineType cur_ln_type = it->lane_marker_type();
      ptr_rm_sm_msg->right_line[0].rm_egorgtlne_type =
          static_cast<uint8_t>(cur_ln_type);
      // FIXME rm_rgtlnesnsrsource shall be uint16
      ptr_rm_sm_msg->right_line[0].rm_rgtlnesnsrsource = static_cast<uint8_t>(
          it->mutable_front_lane_marker_segment()->front().sensor_source());
      ptr_rm_sm_msg->rm_dsblrsn[3] = static_cast<uint8_t>(
          output::MeasurementStateRemap(it->mutable_front_lane_marker_segment()
                                            ->front()
                                            .mutable_geometry()
                                            ->measurement_state(),
                                        cur_ln_type));
      // static_cast<uint32_t>(it->mutable_front_lane_marker_segment()
      //                           ->front()
      //                           .mutable_geometry()
      //                           ->measurement_state());
    }
    if (ll_line_id == it->lane_marker_id()) {
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_id = ll_line_id;
      rm::PolynomLine *ll_line_curve_ptr =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->mutable_curve();
      rm::LineGeometry *ll_line_geometry_ptr =
          it->mutable_front_lane_marker_segment()->front().mutable_geometry();
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].c0_llne =
          ll_line_curve_ptr->Coef(0);
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].c1_llne =
          ll_line_curve_ptr->Coef(1);
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].c2_llne =
          ll_line_curve_ptr->Coef(2);
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].c3_llne =
          ll_line_curve_ptr->Coef(3);
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].start_rnge_llne =
          ll_line_curve_ptr->start_range();
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].end_rnge_llne =
          ll_line_curve_ptr->end_range();
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].probability_llne =
          ll_line_geometry_ptr->exist_probability();
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_geometry[0].valid_llne =
          ll_line_geometry_ptr->valid();
      // ptr_rm_sm_msg->left_line[1].rm_egolftlne_color =
      //     static_cast<uint8_t>(it->mutable_front_lane_marker_segment()->front().color());
      // ptr_rm_sm_msg->left_line[1].rm_egolftlne_type =
      //     static_cast<uint8_t>(it->mutable_front_lane_marker_segment()->front().type());
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_color =
          static_cast<uint8_t>(it->lane_marker_color());
      ptr_rm_sm_msg->left_line[1].rm_egolftlne_type =
          static_cast<uint8_t>(it->lane_marker_type());
      // FIXME rm_lftlnesnsrsource shall be uint16
      ptr_rm_sm_msg->left_line[1].rm_lftlnesnsrsource = static_cast<uint8_t>(
          it->mutable_front_lane_marker_segment()->front().sensor_source());
    }
    if (rr_line_id == it->lane_marker_id()) {
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_id = rr_line_id;
      rm::PolynomLine *rr_line_curve_ptr =
          it->mutable_front_lane_marker_segment()
              ->front()
              .mutable_geometry()
              ->mutable_curve();
      rm::LineGeometry *rr_line_geometry_ptr =
          it->mutable_front_lane_marker_segment()->front().mutable_geometry();
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].c0_rlne =
          rr_line_curve_ptr->Coef(0);
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].c1_rlne =
          rr_line_curve_ptr->Coef(1);
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].c2_rlne =
          rr_line_curve_ptr->Coef(2);
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].c3_rlne =
          rr_line_curve_ptr->Coef(3);
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].start_rnge_rlne =
          rr_line_curve_ptr->start_range();
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].end_rnge_rlne =
          rr_line_curve_ptr->end_range();
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].probability_rlne =
          rr_line_geometry_ptr->exist_probability();
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_geometry[0].valid_rlne =
          rr_line_geometry_ptr->valid();
      // ptr_rm_sm_msg->right_line[1].rm_egorgtlne_color =
      //     static_cast<uint8_t>(it->mutable_front_lane_marker_segment()->front().color());
      // ptr_rm_sm_msg->right_line[1].rm_egorgtlne_type =
      //     static_cast<uint8_t>(it->mutable_front_lane_marker_segment()->front().type());
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_color =
          static_cast<uint8_t>(it->lane_marker_color());
      ptr_rm_sm_msg->right_line[1].rm_egorgtlne_type =
          static_cast<uint8_t>(it->lane_marker_type());
      // FIXME rm_rgtlnesnsrsource shall be uint16
      ptr_rm_sm_msg->right_line[1].rm_rgtlnesnsrsource = static_cast<uint8_t>(
          it->mutable_front_lane_marker_segment()->front().sensor_source());
    }
  }
  rm::ControlWayPoints *way_points_ptr =
      frame_->road_info_provider()->mutable_way_points();
  ptr_rm_sm_msg->control_way_point_target_road_type =
      static_cast<uint8_t>(way_points_ptr->control_way_point_target_roadtype());
  ptr_rm_sm_msg->remain_distance_major_path =
      way_points_ptr->remain_distance_major_path();
  ptr_rm_sm_msg->remain_distance_major_path_type =
      static_cast<uint8_t>(way_points_ptr->remain_distance_major_path_type());
  ptr_rm_sm_msg->reminder_way_point_targe_road_type =
      static_cast<uint8_t>(way_points_ptr->reminder_way_point()
                               .reminder_way_point_target_road_type());
  ptr_rm_sm_msg->reminder_way_point_offset =
      way_points_ptr->reminder_way_point().reminder_way_point_offset();
  ptr_rm_sm_msg->inavlts = way_points_ptr->navi_sts();
  rm::output::EHRSpeedLimit2SMSpeedLimit(
      ptr_rm_sm_msg, way_points_ptr,
      frame_->FindLanePtrFromLanePosition(LanePosition::EGO_LANE));
  ptr_rm_sm_msg->camr_line_mapping =
      way_points_ptr->camr_line_mapping() &&
      road_info_provider_->traffic_road_sensor_status() ==
          SensorStatus::FVCM_AND_HDM_VALID;
  ptr_rm_sm_msg->location_qulity = way_points_ptr->location_quality();
  ptr_rm_sm_msg->rm_dsblrsn[4] =
      static_cast<uint32_t>(way_points_ptr->is_out_odd_range());
  if (way_points_ptr->control_way_point_target_roadtype() ==
          rm::TargetRoadType::LEFT2MAIN ||
      way_points_ptr->control_way_point_target_roadtype() ==
          rm::TargetRoadType::RIGHT2MAIN) {
    ptr_rm_sm_msg->enter_ramp_dist = 0.0F;
    ptr_rm_sm_msg->merge_main_road_dist =
        way_points_ptr->control_way_point_offset();
  } else if (way_points_ptr->control_way_point_target_roadtype() ==
                 rm::TargetRoadType::LEFT2RAMP ||
             way_points_ptr->control_way_point_target_roadtype() ==
                 rm::TargetRoadType::RIGHT2RAMP) {
    ptr_rm_sm_msg->enter_ramp_dist = way_points_ptr->control_way_point_offset();
    ptr_rm_sm_msg->merge_main_road_dist = 0.0F;
  } else {
    ptr_rm_sm_msg->enter_ramp_dist = 0.0F;
    ptr_rm_sm_msg->merge_main_road_dist = 0.0F;
  }
  rm::output::ILCTargetList2RMILCTarget(
      ptr_rm_sm_msg, frame_->mutable_ilc_info()->ilc_target_id_list());

  ptr_rm_sm_msg->rm_dsblrsn.at(1) =
      static_cast<uint32_t>(frame_->danger_pedestrain_id());

  // 3.存储object list及targets info信息
  for (auto &obstacle : *frame_->road_info_provider()->mutable_obstacles()) {
    dpc::Object object;
    output::Obstacle2Object(obstacle.second, object);
    ptr_roadinfo->object_list.emplace_back(std::move(object));
    // meanwhile check the object target property
    switch (obstacle.second.trplontype()) {
    case TargetType::IV:
      output::FrontObstacle2Target(obstacle.second, ptr_targetsinfo->iv);
      tgt_existence_array_[0] = true;
      break;
    case TargetType::IIV:
      output::FrontObstacle2Target(obstacle.second, ptr_targetsinfo->iiv);
      tgt_existence_array_[1] = true;
      break;
    case TargetType::RIV:
      output::RearObstacle2Target(obstacle.second, ptr_targetsinfo->riv);
      tgt_existence_array_[2] = true;
      break;
    case TargetType::FLNIV:
      output::FrontObstacle2Target(obstacle.second, ptr_targetsinfo->flniv);
      tgt_existence_array_[3] = true;
      break;
    case TargetType::FLNIIV:
      output::FrontObstacle2Target(obstacle.second, ptr_targetsinfo->flniiv);
      tgt_existence_array_[4] = true;
      break;
    case TargetType::FLNRIV:
      output::RearObstacle2Target(obstacle.second, ptr_targetsinfo->flnriv);
      tgt_existence_array_[5] = true;
      break;
    case TargetType::SLNIV:
      output::FrontObstacle2Target(obstacle.second, ptr_targetsinfo->slniv);
      tgt_existence_array_[6] = true;
      break;
    case TargetType::SLNIIV:
      output::FrontObstacle2Target(obstacle.second, ptr_targetsinfo->slniiv);
      tgt_existence_array_[7] = true;
      break;
    case TargetType::SLNRIV:
      output::RearObstacle2Target(obstacle.second, ptr_targetsinfo->slnriv);
      tgt_existence_array_[8] = true;
      break;
    case TargetType::OTHER:
      break;
    }
  }
  // after iterating over all the objects, some targets would not be found
  if (tgt_existence_array_[0] == false)
    output::DefaultFrontTarget(ptr_targetsinfo->iv);
  if (tgt_existence_array_[1] == false)
    output::DefaultFrontTarget(ptr_targetsinfo->iiv);
  if (tgt_existence_array_[2] == false)
    output::DefaultRearTarget(ptr_targetsinfo->riv);
  if (tgt_existence_array_[3] == false)
    output::DefaultFrontTarget(ptr_targetsinfo->flniv);
  if (tgt_existence_array_[4] == false)
    output::DefaultFrontTarget(ptr_targetsinfo->flniiv);
  if (tgt_existence_array_[5] == false)
    output::DefaultRearTarget(ptr_targetsinfo->flnriv);
  if (tgt_existence_array_[6] == false)
    output::DefaultFrontTarget(ptr_targetsinfo->slniv);
  if (tgt_existence_array_[7] == false)
    output::DefaultFrontTarget(ptr_targetsinfo->slniiv);
  if (tgt_existence_array_[8] == false)
    output::DefaultRearTarget(ptr_targetsinfo->slnriv);

  tgt_existence_array_ = {
      false, false, false, false, false,
      false, false, false, false}; // reset target existence flag

  for (auto road_edge_it = road_info_provider_->mutable_road_edges()->begin();
       road_edge_it != road_info_provider_->mutable_road_edges()->end();
       road_edge_it++) {
    dpc::RoadEdge pubs_road_edge;
    output::RoadEdge2RoadEdge(*road_edge_it, pubs_road_edge);
    ptr_roadinfo->road_edge_list.emplace_back(std::move(pubs_road_edge));
  }
  // TODO add process to deal with the case with no road edge
  // left road edge
  // FIXME temp solution in the case that the num of road edge is less than 2
  if (road_info_provider_->mutable_road_edges()->size() < 2)
    return Status::OK();
  RoadEdgeSegment left_road_edge_seg;
  road_info_provider_->mutable_road_edges()
      ->at(0)
      .get_current_road_edge_segment(left_road_edge_seg);
  rm::PolynomLine *left_road_edge_curve_ptr =
      left_road_edge_seg.mutable_geometry()->mutable_curve();
  ptr_targetsinfo->left_road_edge.road_edge_c0 =
      left_road_edge_curve_ptr->Coef(0);
  ptr_targetsinfo->left_road_edge.road_edge_c1 =
      left_road_edge_curve_ptr->Coef(1);
  ptr_targetsinfo->left_road_edge.road_edge_c2 =
      left_road_edge_curve_ptr->Coef(2);
  ptr_targetsinfo->left_road_edge.road_edge_c3 =
      left_road_edge_curve_ptr->Coef(3);
  ptr_targetsinfo->left_road_edge.road_edge_existnc_problty =
      left_road_edge_seg.geometry().exist_probability();
  ptr_targetsinfo->left_road_edge.road_edge_viewrange =
      left_road_edge_curve_ptr->end_range();
  ptr_targetsinfo->left_road_edge.road_edge_typ =
      output::RoadEdgeType2TRPRoadEdgeType(left_road_edge_seg.type());
  // right road edge
  RoadEdgeSegment right_road_edge_seg;
  road_info_provider_->mutable_road_edges()
      ->at(1)
      .get_current_road_edge_segment(right_road_edge_seg);
  rm::PolynomLine *right_road_edge_curve_ptr =
      right_road_edge_seg.mutable_geometry()->mutable_curve();
  ptr_targetsinfo->right_road_edge.road_edge_c0 =
      right_road_edge_curve_ptr->Coef(0);
  ptr_targetsinfo->right_road_edge.road_edge_c1 =
      right_road_edge_curve_ptr->Coef(1);
  ptr_targetsinfo->right_road_edge.road_edge_c2 =
      right_road_edge_curve_ptr->Coef(2);
  ptr_targetsinfo->right_road_edge.road_edge_c3 =
      right_road_edge_curve_ptr->Coef(3);
  ptr_targetsinfo->right_road_edge.road_edge_existnc_problty =
      right_road_edge_seg.geometry().exist_probability();
  ptr_targetsinfo->right_road_edge.road_edge_viewrange =
      right_road_edge_curve_ptr->end_range();
  ptr_targetsinfo->right_road_edge.road_edge_typ =
      output::RoadEdgeType2TRPRoadEdgeType(right_road_edge_seg.type());

  // trplon information
  // ptr_targetsinfo->uint8_reserve.at(0) =
  //     static_cast<uint8_t>(road_info_provider_->traffic_road_sensor_status());
  // ptr_targetsinfo->uint8_reserve.at(1) =
  //     static_cast<uint8_t>(road_info_provider_->ehr_current_road_type());
  // ptr_targetsinfo->uint8_reserve.at(2) =
  //     static_cast<uint8_t>(road_info_provider_->desired_lane_id());

  return Status::OK();
}

RoadInfoProvider *RoadModel::mutable_road_info_provider() {
  return road_info_provider_.get();
}

Frame *RoadModel::mutable_frame() { return frame_.get(); }

// uint16_t RoadModel::RunCounter() {
//   counters_++;
//   counters_ = counters_ % 64;
//   return counters_;
// }

common::Status RoadModel::PostProcess() {
  // store the traffic road sensor status of this cycle
  *injector_->mutable_traffic_road_sensor_source_last_cycle() =
      road_info_provider_->traffic_road_sensor_status();
  *injector_->mutable_left_lane_id_last_cycle() =
      frame_->FindLaneIdFromLanePosition(LanePosition::LEFT1);
  *injector_->mutable_right_lane_id_last_cycle() =
      frame_->FindLaneIdFromLanePosition(LanePosition::RIGHT1);
  // // reset and delete in corridor debouncer corresponded to corridor
  // for (auto &&obstacle_iter :
  //      *(frame_->road_info_provider()->mutable_obstacles())) {
  //   obstacle_iter.second.DeleteandReset_incorridor_map();
  //   obstacle_iter.second.reset_trplontype();
  // }
  // reset lane marker id
  for (auto iter = injector_->mutable_lane_marker_id_info_last_cycle()->begin();
       iter != injector_->mutable_lane_marker_id_info_last_cycle()->end();) {
    if (!iter->updated) {
      iter = injector_->mutable_lane_marker_id_info_last_cycle()->erase(iter);
    } else {
      iter->updated = false;
      iter++;
    }
  }
  // reset corridor id
  for (auto iter = injector_->mutable_line_follow_corridor_id_info_last_cycle()
                       ->begin();
       iter !=
       injector_->mutable_line_follow_corridor_id_info_last_cycle()->end();) {
    if (!iter->second) {
      iter =
          injector_->mutable_line_follow_corridor_id_info_last_cycle()->erase(
              iter);
    } else {
      iter->second = false;
      iter++;
    }
  }
  // reassign the object
  // if (frame_->mutable_lane_info()->size() > 1U &&
  //     frame_->mutable_lane_info()->front().lane().lane_position() ==
  //         LanePosition::EGO_LANE) {
  //   // if there is more than 1 lane and ego lane exists
  //   for (auto obj_it = road_info_provider_->mutable_obstacles()->begin();
  //        obj_it != road_info_provider_->mutable_obstacles()->end(); obj_it++)
  //        {
  //     if (!obj_it->second.is_already_assigned() &&
  //         obj_it->second.prv_corridor_id() != 0U) {
  //       // object is not been assigned to any corridor and was assigned to
  //       // corridor in previous cycle
  //       // get the corresponding lane id to which the object was assiged
  //       before obj_it->second.Add_duration_of_intermidiate_status();
  //       // accout the duration of not-being selected
  //       if (obj_it->second.duration_of_intermidiate_status() > 0.16)
  //         continue;
  //       // continue if the duration is greater than 0.16
  //       uint8_t prv_lane_id = std::floor(obj_it->second.prv_corridor_id() /
  //       10);
  //       // find the corresponding lane
  //       auto lane_iter = std::find_if(
  //           frame_->mutable_lane_info()->begin(),
  //           frame_->mutable_lane_info()->end(), [&](LaneInfo &lane) {
  //             return lane.mutable_lane()->lane_id() == prv_lane_id;
  //           });
  //       if (lane_iter != frame_->mutable_lane_info()->end() &&
  //           !lane_iter->mutable_corridor()->empty()) {
  //         // if the lane exists and the corridor list is not empty
  //         // find the object location affilated to this corridor
  //         auto loc_iter =
  //             std::find_if(lane_iter->mutable_corridor()
  //                              ->front()
  //                              .mutable_object_location()
  //                              ->begin(),
  //                          lane_iter->mutable_corridor()
  //                              ->front()
  //                              .mutable_object_location()
  //                              ->end(),
  //                          [&](ObjectLocation &loc) {
  //                            return loc.object_id == obj_it->second.id();
  //                          });
  //         // change the in_corridor attribute
  //         loc_iter->in_corridor = true;
  //       }
  //     }
  //   }
  // }

  // ats switch decider
  frame_->set_ats_switch_flag(ats_switch_decider_.Run(frame_.get()));
  frame_->set_ats_switch_state(ats_switch_decider_.state());

  // targets info selection
  auto ego_lane_info_iter = std::find_if(
      frame_->mutable_lane_info()->begin(), frame_->mutable_lane_info()->end(),
      [](rm::LaneInfo &lane_info) {
        return lane_info.lane().lane_position() == LanePosition::EGO_LANE;
      });
  std::map<uint32_t, Obstacle> *obstacles =
      frame_->road_info_provider()->mutable_obstacles();
  if (ego_lane_info_iter != frame_->mutable_lane_info()->end() &&
      !ego_lane_info_iter->mutable_corridor()->empty()) {
    std::vector<Corridor> *corridor_list_ptr =
        ego_lane_info_iter->mutable_corridor();

    // it is now assumed that the corridor list only contains 1 element
    std::pair<std::list<ObjectLocation>, std::list<ObjectLocation>>
        temp_in_corridor_list;
    std::pair<std::list<ObjectLocation>, std::list<ObjectLocation>>
        temp_left_corridor_list;
    std::pair<std::list<ObjectLocation>, std::list<ObjectLocation>>
        temp_right_corridor_list;
    std::list<ObjectLocation> front_pedestrian_corridor_list;
    std::list<ObjectLocation> *in_corridor_list_ptr;
    std::list<ObjectLocation> *left_corridor_list_ptr;
    std::list<ObjectLocation> *right_corridor_list_ptr;
    double lateral_dis_constraints = std::max(
        1.5 * ego_lane_info_iter->mutable_lane()->lane_width(),
        3.5); // take the width of the first lane segement as constriant
    for (auto iter =
             corridor_list_ptr->begin()->mutable_object_location()->begin();
         iter != corridor_list_ptr->begin()->mutable_object_location()->end();
         iter++) {
      // front or back object
      if (iter->classification == ObstacleType::PEDESTRIAN) {
        if (iter->in_corridor == true &&
            iter->reference_point2corridor_frenet.s() > 0.0) {
          front_pedestrian_corridor_list.emplace_back(*iter);
        } else {
          // do nothing
        }
        // continue;
      }

      if (iter->reference_point2corridor_frenet.s() > 0) {
        in_corridor_list_ptr = &temp_in_corridor_list.first;
        left_corridor_list_ptr = &temp_left_corridor_list.first;
        right_corridor_list_ptr = &temp_right_corridor_list.first;
      } else {
        in_corridor_list_ptr = &temp_in_corridor_list.second;
        left_corridor_list_ptr = &temp_left_corridor_list.second;
        right_corridor_list_ptr = &temp_right_corridor_list.second;
      }

      if (iter->in_corridor == true) {
        in_corridor_list_ptr->emplace_back(*iter);
        auto obj_iter = obstacles->find(iter->object_id);
        if (obj_iter->second.Generate_virtual_corridor(
                ego_lane_info_iter->mutable_boundary())) {
          corridor_list_ptr->emplace_back(obj_iter->second.virtual_corridor());
        } else {
          /*do nothing*/
        }
      } else {
        if (iter->reference_point2corridor_frenet.l() <=
                lateral_dis_constraints &&
            iter->reference_point2corridor_frenet.l() >=
                0) { // object on the front left side of ego corridor{
          left_corridor_list_ptr->emplace_back(*iter);
        } else if (iter->reference_point2corridor_frenet.l() >=
                       -lateral_dis_constraints &&
                   iter->reference_point2corridor_frenet.l() <
                       0) { // object on the front right side of ego corridor{
          right_corridor_list_ptr->emplace_back(*iter);
        }
      }
    }
    // set the trplontype of the object which fullfills the long/lat distance
    // conditionsk
    bool disable_iiv_selection, disable_siiv_selection, disable_fiiv_selection;
    double iv_frenet_s_sub_by_half_length = 1000.0;
    // IV
    if (!temp_in_corridor_list.first.empty()) {
      auto iv_iter = std::min_element(temp_in_corridor_list.first.begin(),
                                      temp_in_corridor_list.first.end());
      auto obstacle_iv_iter = obstacles->find(iv_iter->object_id);
      if (obstacle_iv_iter->second.is_in_long_distance_range()) {
        obstacle_iv_iter->second.set_trplontype(TargetType::IV);
        iv_frenet_s_sub_by_half_length =
            iv_iter->reference_point2corridor_frenet.s() -
            0.5 * iv_iter->size_length;
        temp_in_corridor_list.first.erase(
            iv_iter); //  erase the minima from in corridor list
        disable_iiv_selection = false;
      } else {
        disable_iiv_selection = true;
      }
    }
    // IIV
    if (!temp_in_corridor_list.first.empty() &&
        disable_iiv_selection == false) {
      auto iiv_iter = std::min_element(temp_in_corridor_list.first.begin(),
                                       temp_in_corridor_list.first.end());
      auto obstacle_iiv_iter = obstacles->find(iiv_iter->object_id);
      if (obstacle_iiv_iter->second.is_in_long_distance_range()) {
        obstacle_iiv_iter->second.set_trplontype(TargetType::IIV);
      }
    }
    // RIV
    if (!temp_in_corridor_list.second.empty()) {
      auto riv_iter = std::max_element(temp_in_corridor_list.second.begin(),
                                       temp_in_corridor_list.second.end());
      obstacles->find(riv_iter->object_id)
          ->second.set_trplontype(TargetType::RIV);
    }
    // FLNIV
    if (!temp_left_corridor_list.first.empty()) {
      auto flniv_iter = std::min_element(temp_left_corridor_list.first.begin(),
                                         temp_left_corridor_list.first.end());
      auto obstacle_flniv_iter = obstacles->find(flniv_iter->object_id);
      if (obstacle_flniv_iter->second.is_in_long_distance_range()) {
        obstacle_flniv_iter->second.set_trplontype(TargetType::FLNIV);
        temp_left_corridor_list.first.erase(flniv_iter);
        disable_fiiv_selection = false;
        obstacle_flniv_iter->second.Run_lane_change_intention_detector(
            *flniv_iter);
      } else {
        disable_fiiv_selection = true;
      }
    }
    // FLNIIV
    if (!temp_left_corridor_list.first.empty() &&
        disable_fiiv_selection == false) {
      auto flniiv_iter = std::min_element(temp_left_corridor_list.first.begin(),
                                          temp_left_corridor_list.first.end());
      auto obstacle_flniiv_iter = obstacles->find(flniiv_iter->object_id);
      if (obstacle_flniiv_iter->second.is_in_long_distance_range()) {
        obstacle_flniiv_iter->second.set_trplontype(TargetType::FLNIIV);
        obstacle_flniiv_iter->second.Run_lane_change_intention_detector(
            *flniiv_iter);
      }
    }
    // FLNRIV
    if (!temp_left_corridor_list.second.empty()) {
      auto flnriv_iter =
          std::max_element(temp_left_corridor_list.second.begin(),
                           temp_left_corridor_list.second.end());
      obstacles->find(flnriv_iter->object_id)
          ->second.set_trplontype(TargetType::FLNRIV);
    }
    // SLNIV
    if (!temp_right_corridor_list.first.empty()) {
      auto slniv_iter = std::min_element(temp_right_corridor_list.first.begin(),
                                         temp_right_corridor_list.first.end());
      auto obstacle_slniv_iter = obstacles->find(slniv_iter->object_id);
      if (obstacle_slniv_iter->second.is_in_long_distance_range()) {
        obstacle_slniv_iter->second.set_trplontype(TargetType::SLNIV);
        temp_right_corridor_list.first.erase(slniv_iter);
        disable_siiv_selection = false;
        obstacle_slniv_iter->second.Run_lane_change_intention_detector(
            *slniv_iter);
      } else {
        disable_siiv_selection = true;
      }
    }
    // SLNIIV
    if (!temp_right_corridor_list.first.empty() &&
        disable_siiv_selection == false) {
      auto slniiv_iter =
          std::min_element(temp_right_corridor_list.first.begin(),
                           temp_right_corridor_list.first.end());
      auto obstacle_slniiv_iter = obstacles->find(slniiv_iter->object_id);
      if (obstacle_slniiv_iter->second.is_in_long_distance_range()) {
        obstacle_slniiv_iter->second.set_trplontype(TargetType::SLNIIV);
        obstacle_slniiv_iter->second.Run_lane_change_intention_detector(
            *slniiv_iter);
      }
    }
    // SLNRIV
    if (!temp_right_corridor_list.second.empty()) {
      auto slnriv_iter =
          std::max_element(temp_right_corridor_list.second.begin(),
                           temp_right_corridor_list.second.end());
      obstacles->find(slnriv_iter->object_id)
          ->second.set_trplontype(TargetType::SLNRIV);
    }

    // ped iv
    if (!front_pedestrian_corridor_list.empty()) {
      auto pediv_iter = std::min_element(front_pedestrian_corridor_list.begin(),
                                         front_pedestrian_corridor_list.end());
      auto obstacle_pediv_iter = obstacles->find(pediv_iter->object_id);
      if (obstacle_pediv_iter->second.is_in_long_distance_range() &&
          iv_frenet_s_sub_by_half_length >
              pediv_iter->reference_point2corridor_frenet.s()) {
        frame_->set_danger_pedestrain_id(pediv_iter->object_id);
      }
    }
  }

  // reset and delete in corridor debouncer corresponded to corridor
  for (auto &&obstacle : *(frame_->road_info_provider()->mutable_obstacles())) {
    obstacle.second.DeleteandReset_incorridor_map();
    obstacle.second
        .Reset_vulnerable_object_critical_ego_corridor_debouncer_pair_post();
    obstacle.second.reset_trplontype();
    if (!frame_->is_straight_driving() ||
        !obstacle.second.is_front_adjacent_target()) {
      // if ego veh is not driving straightly or the object is not assigned to
      // be fiv/fiiv/siv/siiv, then reset the lane change intention detector
      obstacle.second.Reset_lane_change_intention_detector();
    }
  }

  return Status::OK();
} // namespace rm

common::Status RoadModel::PreProcess() {
    // run in corridor longitudinal debouncer
  bool ego_corridor_not_exist =
      (frame_->FindLaneIdFromLanePosition(rm::LanePosition::EGO_LANE) == 0U);
  double ego_velocity_kph =
      VehicleStateProvider::Instance()->lon_velocity() * rm::MPS2KPH;
  if (frame_->objects_input_status()) {
    for (auto obstacle_iter =
             frame_->road_info_provider()->mutable_obstacles()->begin();
         obstacle_iter !=
         frame_->road_info_provider()->mutable_obstacles()->end();
         obstacle_iter++) {
      obstacle_iter->second.Run_in_corridor_long_distance_debouncer(
          ego_velocity_kph);
      obstacle_iter->second.unset_is_already_assigned();
      obstacle_iter->second.Check_virtual_corridor_generation_plausibility();
      obstacle_iter->second.set_is_vulnerable_critical_obstacle(false);
      if (ego_corridor_not_exist) {
        obstacle_iter->second.Reset_was_in_ego_corridor_assured_flfp();
        obstacle_iter->second
            .Reset_vulnerable_object_critical_ego_corridor_debouncer_pair_pre();
      }
    }
  }
  
  // straight driving secneario detecting
  bool straight_driving_flag = straight_driving_detector_.RunDetector();
  // std::cout << "straight driving: " << straight_driving_flag << std::endl;
  frame_->set_is_straight_driving(straight_driving_flag);
  
  // sort the lane info vector in the order of position
  // ego lane at first postion, left at second and right at third
  std::array<rm::LanePosition, 3> position_array = {
      LanePosition::EGO_LANE, LanePosition::LEFT1, LanePosition::RIGHT1};
  auto idx_iter = frame_->mutable_lane_info()->begin();
  for (auto pos : position_array) {
    // find the lane with specific postion
    auto lane_info_it = std::find_if(
        frame_->mutable_lane_info()->begin(),
        frame_->mutable_lane_info()->end(), [&](LaneInfo &lane_info) {
          return pos == lane_info.mutable_lane()->lane_position();
        });
    // if found, swap the iterator in the list
    if (lane_info_it != frame_->mutable_lane_info()->end()) {
      std::iter_swap(idx_iter, lane_info_it);
      idx_iter++;
    } else {
      // do nothing
    }
  }
  // run prediction factor wrt ego yaw rate
  Obstacle::Compute_prediction_factor_ego_yaw_rate();
  // in case of traffic road sensor status switches from tja to nrp(vice
  // versa), inherite the previous in-corridor-map for ego/left/right lane
  if ((road_info_provider_->traffic_road_sensor_status() ==
           SensorStatus::FVCM_AND_HDM_VALID &&
       injector_->traffic_road_sensor_source_last_cycle() ==
           SensorStatus::ONLY_FVCM_VALID) ||
      (road_info_provider_->traffic_road_sensor_status() ==
           SensorStatus::ONLY_FVCM_VALID &&
       injector_->traffic_road_sensor_source_last_cycle() ==
           SensorStatus::FVCM_AND_HDM_VALID)) {
    for (auto obj_iter = road_info_provider_->mutable_obstacles()->begin();
         obj_iter != road_info_provider_->mutable_obstacles()->end();
         obj_iter++) {
      obj_iter->second.Update_incorridor_map_key(
          injector_->ego_lane_id_last_cycle(),
          frame_->FindLaneIdFromLanePosition(LanePosition::EGO_LANE));
      obj_iter->second.Update_incorridor_map_key(
          injector_->left_lane_id_last_cycle(),
          frame_->FindLaneIdFromLanePosition(LanePosition::LEFT1));
      obj_iter->second.Update_incorridor_map_key(
          injector_->right_lane_id_last_cycle(),
          frame_->FindLaneIdFromLanePosition(LanePosition::RIGHT1));
    }
  }

  return Status::OK();
}

void RoadModel::ResetInjector() {
  *(injector_->mutable_ego_lane_id_last_cycle()) = static_cast<uint8_t>(0);
  *(injector_->mutable_left_lane_id_last_cycle()) = static_cast<uint8_t>(0);
  *(injector_->mutable_right_lane_id_last_cycle()) = static_cast<uint8_t>(0);
  *(injector_->mutable_traffic_road_sensor_source_last_cycle()) =
      SensorStatus::INIT;
  injector_->mutable_lane_marker_id_info_last_cycle()->clear();
  injector_->mutable_line_follow_corridor_id_info_last_cycle()->clear();
}

void RoadModel::RelayCemInputStatus(
    dpc::RMSMMessage *const ptr_rm_sm_msg) {
  // ptr_rm_sm_msg->rm_header.status =
  //     std::bitset<32>(road_info_provider_->mutable_objects_cem_header()
  //                         ->status()
  //                         .to_ullong())
  //         .to_ulong();
  ptr_rm_sm_msg->rm_header.status = 0;
}

} // namespace rm
